﻿// main.cpp
#include <cassert>
#include <cstdlib>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <iterator>
#include <sstream>
#include <string>
#include <vector>
#include <sys/stat.h>
#include "analysis_cmdline_args.h"

using namespace std;

namespace beditor {

    constexpr auto app_name = "binary_editor";
    constexpr auto version_str = "1.0.3";

    // オプション
    using option_t = unsigned int;
    constexpr option_t option_none {0};          // オプションなし
    constexpr option_t option_lower_case {0b1};  // 小文字で表示する
    static option_t option {option_none};

    // 読み込むサイズ
    static uintmax_t read_size {numeric_limits<uintmax_t>::max()};

	// 出力開始位置
	static uintmax_t begin_pos {0};

	static bool get_file_size(const string& path, uintmax_t* file_size);
    static vector<char> read_file(ifstream& ifs, uintmax_t begin_pos, uintmax_t raed_size);
    static void print_help();
    static void print_version();
    static void on_option_help(const string& param);
    static void on_option_version(const string& param);
    static void on_option_lower(const string& param);
    static void on_option_size(const string& param);
    static void on_option_begin(const string& param);

    // オプション一覧
    static const vector<cmd_args::OptionInfo> options {
        { "-l", "--lower", on_option_lower, false },
        {"-s", "--size", on_option_size, true},
        {"-b", "--begin", on_option_begin, true},
    };

    // 大文字・小文字
    enum class Letter {
        upper,
        lower,
    };


    // バイト列を標準出力へ出力する
    //
    // address: 対象のバイト列を指し示すポインタ
    // size: 対象のバイト列のバイト数
    // letter: 16進数のアルファベットの表示方法
    static void print_byte_string(const void* address, size_t size, Letter letter);
}

int main(int argc, char** argv)
{
    // コマンドライン引数を vector に詰め替える
    auto cmdline_args = cmd_args::argv_to_vector(argc, argv);
    
    // ヘルプとバージョン情報のオプションを登録
    cmd_args::regist_help({"-h", "--help", beditor::on_option_help});
    cmd_args::regist_version({"-v", "--version", beditor::on_option_version});

    // コマンドライン引数の解析
    switch (cmd_args::analysis_cmdline_args(cmdline_args, 2, beditor::options)) {
    case cmd_args::AnalysisCmdlineArgsResult::correct:
        break;

    case cmd_args::AnalysisCmdlineArgsResult::correct_exit:
        quick_exit(EXIT_SUCCESS);
        break;

    case cmd_args::AnalysisCmdlineArgsResult::error_args_not_enough:
        cerr << "コマンドライン引数の指定が正しくありません。\n";
        beditor::print_help();
        quick_exit(EXIT_FAILURE);
        break;

    case cmd_args::AnalysisCmdlineArgsResult::error_invalid_option:
        cerr << "不正なオプションが指定されています。\n";
        beditor::print_help();
        quick_exit(EXIT_FAILURE);
        break;

    case cmd_args::AnalysisCmdlineArgsResult::error_invalid_parameter:
        cerr << "オプションのパラメータに不備があります。\n";
        beditor::print_help();
        quick_exit(EXIT_FAILURE);
        break;

    default:
        break;
    }

	// ファイルサイズを取得
    string path {cmdline_args.at(1)};
	uintmax_t file_size {};
	if (beditor::get_file_size(path, &file_size) == false) {
		cerr << path << "のサイズ取得に失敗しました。\n";
        quick_exit(EXIT_FAILURE);
	}

	// ファイルサイズを越えないように調整する
	if (beditor::begin_pos > file_size) {
		beditor::begin_pos = file_size;
	}
	if (beditor::begin_pos + beditor::read_size > file_size) {
		beditor::read_size = file_size - beditor::begin_pos;
	}

    // 対象のファイルをオープン
    ifstream ifs(path, ios_base::in | ios_base::binary);
    if (!ifs) {
        cerr << path << "が開けません。\n";
        quick_exit(EXIT_FAILURE);
    }

	// ファイルの内容を読み込む
	vector<char> bytes {beditor::read_file(ifs, beditor::begin_pos, beditor::read_size)};

	// バイト列を出力
	beditor::print_byte_string(
		bytes.data(),
		bytes.size(),
		(beditor::option & beditor::option_lower_case) ? beditor::Letter::lower : beditor::Letter::upper
	);
}

namespace beditor {

	// ファイルサイズを返す
	static bool get_file_size(const string& path, uintmax_t* file_size)
	{
		assert(file_size);

		// ファイルの情報を取得
		struct stat st {};
		if (stat(path.c_str(), &st) != 0) {
			return false;
		}

		// ファイルかどうか
		// S_ISREG(st.st_mode); の方がシンプルだが、Visual Studio では使えない。
		if ((st.st_mode & S_IFMT) != S_IFREG) {
			return false;
		}

		*file_size = st.st_size;
		return true;
	}

	// ファイルから必要なサイズ分だけ読み込む
    static vector<char> read_file(ifstream& ifs, uintmax_t begin_pos, uintmax_t raed_size)
    {
        if (begin_pos == 0 && raed_size == numeric_limits<uintmax_t>::max()) {
            // ファイルの中身をすべて読み込む
            istreambuf_iterator<char> it_ifs_begin(ifs);
            istreambuf_iterator<char> it_ifs_end {};
            vector<char> bytes(it_ifs_begin, it_ifs_end);
            return bytes;
        }
        else {
            // ファイルの中身を必要なサイズ分だけ読み込む。
			vector<char> bytes(static_cast<vector<char>::size_type>(raed_size));
			ifs.seekg(begin_pos, ios_base::beg);
			ifs.read(bytes.data(), raed_size);
			return bytes;
        }
    }

    // ヘルプメッセージを出力
    static void print_help()
    {
        cout << "使い方:\n"
             << "> " << app_name << " 対象のファイルパス\n"
             << "> " << app_name << " 対象のファイルパス オプション\n"
             << "> " << app_name << " -h（または --help)\n"
             << "> " << app_name << " -v（または --version）\n"
             << "\n"
             << "-h（--help）    使い方を表示します。\n"
             << "-v（--version） バージョン情報を表示します。\n"
             << "\n"
             << "オプション:\n"
             << "-l (--lower)      --> 16進数のアルファベット部分を小文字で表示します。\n"
             << "-s (--size) 整数  --> 指定サイズまでで表示を終了します。"
             << "-b (--begin) 整数 --> 指定位置から表示を開始します。"
             << endl;
    }

    // バージョン情報を出力
    static void print_version()
    {
        cout << app_name << " version: " << version_str << endl;
    }

    // ヘルプオプションを発見したときに呼び出される関数
    static void on_option_help(const std::string& param)
    {
        (void)param;
        print_help();
    }

    // バージョン情報オプションを発見したときに呼び出される関数
    static void on_option_version(const std::string& param)
    {
        (void)param;
        print_version();
    }

    // 小文字化オプションを発見したときに呼び出される関数
    static void on_option_lower(const std::string& param)
    {
        (void)param;
        option |= option_lower_case;
    }

    // サイズ指定オプションを発見したときに呼び出される関数
    static void on_option_size(const string& param)
    {
        istringstream iss {param};
        uintmax_t size {};
        iss >> size;
        if (iss) {
            read_size = size;
        }
        else {
            cerr << "サイズの指定が不適切です。\n";
            quick_exit(EXIT_FAILURE);
        }
    }

	// 出力開始位置の指定オプションを発見したときに呼び出される関数
	static void on_option_begin(const string& param)
	{
        istringstream iss {param};
        uintmax_t begin {};
        iss >> begin;
        if (iss) {
            begin_pos = begin;
        }
        else {
            cerr << "出力開始位置の指定が不適切です。\n";
            quick_exit(EXIT_FAILURE);
        }
	}

    // バイト列を標準出力へ出力する
    static void print_byte_string(const void* address, size_t size, Letter letter)
    {
        auto p = static_cast<const unsigned char*>(address);
        auto remain_byte = size;

        while (remain_byte > 0) {
            size_t byte_num {remain_byte >= 16 ? 16 : remain_byte};

            for (size_t i {0}; i < byte_num; ++i) {
                cout << setw(2) << setfill('0') << hex
                     << (letter == Letter::upper ? uppercase : nouppercase)
                     << static_cast<unsigned int>(*p) << " ";
                ++p;
            }
            cout << "\n";

            remain_byte -= byte_num;
        }
    }
}